@@ -3,12 +3,14 @@
  */
 
 /* 
- * Program to assign static device names to SCSI devices as opposed to the 
- * dynamic numbering of the Linux-1.0--2.4 kernel.
- * Program written by Eric Youngdale, <aric@aib.com>
+ * Program to assign static device names to SCSI devices as opposed to the
+ * dynamic numbering of the Linux-1.0 -- 2.4 kernel.
+ * Program written by Eric Youngdale <aric@aib.com>
+ * Reworked 1/2000 Kurt Garloff <garloff@suse.de>
+ * 
+ * Copyright: GNU GPL (***ERIC NEEDS TO ASKED !!! ***)
  *
  * Changes:
- * Version: 1.6
  * * 2000/01/16: Kurt Garloff <garloff@suse.de>: 
  *   - SCSI_DISK_MAJOR
  *   - Fixed lots of warnings
@@ -18,7 +20,25 @@
  *   - inquire: Request takes a few bytes, response buffer is somewhat smaller
  *		use getstr() to get and truncate strings
  *   - comparison: Empty Strings of failed inquiries do not fit
- *
+ * -> Released Version: 1.6
+ * 
+ * * 2000/01/17: Kurt Garloff <garloff@suse.de>:
+ *   Major changes:
+ *   - Now scan completely relying on sg. Does inquiry for every device
+ *     immediately (it's not that expensive), and does interpret the 
+ *     periph dev qualifier field to build the other high-level SCSI devs.
+ *     -> removable devices will work, this way.
+ *     For harddisks, the partitions will be scanned, for tapes the n version
+ *     be created.
+ *   Minor changes:
+ *   - new inquiry () function replacing old inquire and get_version_no
+ *   - SCSI_DISK_MAJOR(x) does not assume only 16 SCSI disks any more
+ *   - quiet flag -q
+ *   - Link alias names flag -L
+ *   - -c maxmiss option: Tolerate maxmiss missing sg devs.
+ *   - rev and hostname fields
+ * -> Released Version: 2.0
+ * 
  */
 
 #include <stdio.h>
@@ -34,11 +54,9 @@
 #include <stdlib.h>
 #include <sys/ioctl.h>
 
-#undef SCSI_DISK_MAJOR
-#define SCSI_DISK_MAJOR SCSI_DISK0_MAJOR
 
-//static char rcsid[] ="$Id: scsidev.c,v 1.3 2000/01/16 22:11:50 garloff Exp $";
-static char *versinfo = "scsidev " VERSION " 2000/01/16";
+//static char rcsid[] ="$Id: scsidev.c,v 1.8.2.3 2000/01/28 12:14:34 garloff Exp $";
+static char *versid = "scsidev " VERSION " 2000/01/16";
 
 #include "config.h"
 
@@ -59,8 +77,11 @@ static char *versinfo = "scsidev " VERSION " 2000/01/16";
 #include <getopt.h>
 
 int use_symlink = 0;
+int symlink_alias = 0;
 int filemode = 0600;
 int verbose = 0;
+int quiet = 0;
+int maxmiss = 0;
 char * no_serial = "No serial number";
 
 #define DEVSCSI "/dev/scsi"
@@ -88,13 +109,16 @@ struct regnames
   char * name;
   char * manufacturer;
   char * model;
-  char * rev;
   char * serial;
-  struct regnames * alias;
+  char * rev;
+  char inq_devtp;
+  char rmvbl;
+  struct regnames * alias; // TBR
   int major;
   int minor;
   int hostnum;
   int hostid;
+  char * hostname;
   int chan;
   int id;
   int lun;
@@ -103,13 +127,42 @@ struct regnames
 
 struct regnames * reglist = NULL;
 
+void build_special();
+int inquiry (struct regnames *);
+
+
 /*
  * Used to maintain a list of the nodes that we have seen
  * which we know to be OK and active.
  */
-void register_dev(char * name, int major, int minor, int hnum, 
-	     int hid, int chan, int id, 
-	     int lun, int part, struct regnames * alias)
+
+void dumplist ()
+{
+	struct regnames * pnt;
+	for (pnt = reglist; pnt; pnt = pnt->next) 
+	{
+		printf ("%s: %s %s %s (%s)\n", pnt->name,
+			pnt->manufacturer, pnt->model, pnt->rev, pnt->serial);
+		printf ("  on %s (%d-%x): c%di%dl%d", pnt->hostname,
+			pnt->hostnum, pnt->hostid, pnt->chan, pnt->id, pnt->lun);
+		if (pnt->partition != -1) printf ("p%d", pnt->partition);
+		printf (" %c %x-%x\n", (SCSI_BLK_MAJOR(pnt->major)? 'b': 'c'),
+			pnt->major, pnt->minor);
+	}
+}
+
+struct regnames * regnames_dup (struct regnames * rpnt)
+{
+	struct regnames * rpnt1 = malloc (sizeof (struct regnames));
+	memcpy (rpnt1, rpnt, sizeof (struct regnames));
+	//rpnt1->next = reglist; reglist = rpnt1;
+	return rpnt1;
+}
+
+struct regnames * register_dev (char * name, int major, int minor, 
+				int hnum, int hid, int chan, int id,
+				int lun, int part, char * hostname, 
+				struct regnames * alias)
 {
   struct regnames * rpnt;
   char * pnt;
@@ -125,13 +178,15 @@ void register_dev(char * name, int major, int minor, int hnum,
   rpnt->id = id;
   rpnt->lun = lun;
   rpnt->partition = part;
-  rpnt->next = reglist;
+  if (hostname) rpnt->hostname = strdup (hostname); 
+  else rpnt->hostname = 0;
   rpnt->alias = alias;
   /*
    * Initialize this - they may be needed later.
    */
   rpnt->model = rpnt->manufacturer = rpnt->serial = NULL;
-  reglist = rpnt;
+  rpnt->next = reglist; reglist = rpnt;
+  return rpnt;
 }
 
 /*
@@ -171,11 +226,11 @@ void sanitize_sdev()
 	      break;
 	    }
 	}
-      if( rpnt == NULL )
+      if (rpnt == NULL)
 	{
-	  strcpy(filename, DEVSCSI); strcat(filename, "/");
-	  strcat(filename, de->d_name);
-	  status = stat(filename, &statbuf);
+	  strcpy (filename, DEVSCSI); strcat (filename, "/");
+	  strcat (filename, de->d_name);
+	  status = stat (filename, &statbuf);
 	  if(    status == 0
 	      && (S_ISCHR(statbuf.st_mode) || S_ISBLK(statbuf.st_mode)))
 	    {
@@ -184,17 +239,16 @@ void sanitize_sdev()
 	       * with.  No big deal, stat it so we get the particulars, then
 	       * create a new one with a safe minor number.
 	       */
-	      unlink(filename);
-
-	      status = mknod(filename, statbuf.st_mode ,
-			     makedev(major(statbuf.st_rdev),255));
-	      status = chmod(filename, statbuf.st_mode);
-	      chown(filename, statbuf.st_uid, statbuf.st_gid);
+	      unlink (filename);
 
+	      status = mknod (filename, statbuf.st_mode ,
+			      makedev(major(statbuf.st_rdev),255));
+	      status = chmod (filename, statbuf.st_mode);
+	      chown (filename, statbuf.st_uid, statbuf.st_gid);
 	    }
 	}
     }
-  closedir(sdir);
+  closedir (sdir);
 
 }
 
@@ -209,18 +263,18 @@ void flush_sdev()
    * The idea is that if we have added/removed devices, the numbers might have
    * changed.
    */
-  sdir = opendir(DEVSCSI);
-  while(1==1)
+  sdir = opendir (DEVSCSI);
+  while (1==1)
     {
       de = readdir(sdir);
-      if(de == NULL) break;
-      if(de->d_name[0] != 's' && de->d_name[0] != 'r') continue;
-      strcpy(filename, DEVSCSI); strcat (filename, "/");
-      strcat(filename, de->d_name);
-      unlink(filename);
+      if (de == NULL) break;
+      if (de->d_name[0] != 's' && de->d_name[0] != 'n') continue;
+      strcpy (filename, DEVSCSI); strcat (filename, "/");
+      strcat (filename, de->d_name);
+      unlink (filename);
     }
-  closedir(sdir);
-  printf("Flushed old " DEVSCSI " entries...\n");
+  closedir (sdir);
+  if (!quiet) printf ("Flushed old " DEVSCSI " entries...\n");
 
 }
 
@@ -229,7 +283,7 @@ void flush_sdev()
  * if it does make sure the major and minor numbers are correct
  * and save permissions and ownerships if this is the case.
  */
-void update_device(char * path, int devtype, int major, int minor)
+void update_device (char * path, int devtype, int major, int minor)
 {
   struct stat statbuf;
   int recreate;
@@ -248,8 +302,8 @@ void update_device(char * path, int devtype, int major, int minor)
       /*
        * Make sure we have the correct device type too.
        */
-      if( (statbuf.st_mode & S_IFMT) != devtype ) {
-	  unlink(path);
+      if ( (statbuf.st_mode & S_IFMT) != devtype ) {
+	  unlink (path);
 	  recreate = 1;
 	}
       /*
@@ -261,7 +315,8 @@ void update_device(char * path, int devtype, int major, int minor)
 	  newmode = statbuf.st_mode;
 	  uid = statbuf.st_uid;
 	  gid = statbuf.st_gid;
-	  unlink(path);
+	  unlink (path);
+	  if (verbose == 2) printf ("mismatch: Recreate %s (%i-%i)\n", path, major, minor);
 	  recreate = 1;
 	}
     }
@@ -273,9 +328,10 @@ void update_device(char * path, int devtype, int major, int minor)
     {
       status = mknod(path, newmode,
 		     makedev(major,minor));
+      //printf("Recreate maj %i min %i\n", major, minor);
       if( status == -1 ) {
-	  fprintf(stderr,"mknod (%s) failed\n", path);
-	  exit(1);
+	  fprintf (stderr, "mknod (%s) failed\n", path);
+	  exit (1);
 	}
       
       /*
@@ -285,8 +341,8 @@ void update_device(char * path, int devtype, int major, int minor)
        */
       status = chmod(path, newmode);
       if( status == -1 ) {
-	  fprintf(stderr,"chmod (%s) failed\n", path);
-	  exit(1);
+	  fprintf (stderr, "chmod (%s) failed\n", path);
+	  exit (1);
 	}
       
       if( uid != -1 )
@@ -294,252 +350,292 @@ void update_device(char * path, int devtype, int major, int minor)
     }  
 }
 
-void build_sdev(int major, int mode, int devtype, char devchar)
+void create_dev (struct regnames *rpnt)
 {
-  int chan;
-  int fd;
-  char hostname[60];
-  int newmode;
-  //int recreate;
-  int h_id;
-  int id[2];
-  int lun;
-  int minor = 0;
-  //int uid, gid;
-  struct stat statbuf;
-  int scsi_id;
-  char scsidev[64];
-  char linkto[64];
-  int status;
+	char linkto[64];
+	int status;
+	int newmode;
+	int major = rpnt->major, minor = rpnt->minor;
+	char scsidev[128];
+	struct stat statbuf;
+	int devtype = (SCSI_BLK_MAJOR (major)? S_IFBLK: S_IFCHR);
+	strcpy (scsidev, DEVSCSI); strcat (scsidev, "/"); 
+	strcat (scsidev, rpnt->name);
+	
+	if (use_symlink)
+	{
+		switch(major)
+		{
+		 case SCSI_DISK0_MAJOR:
+		 case SCSI_DISK1_MAJOR:
+		 case SCSI_DISK2_MAJOR:
+		 case SCSI_DISK3_MAJOR:
+		 case SCSI_DISK4_MAJOR:
+		 case SCSI_DISK5_MAJOR:
+		 case SCSI_DISK6_MAJOR:
+		 case SCSI_DISK7_MAJOR:
+			if (minor & 0xf)
+				sprintf (linkto, "/dev/sd%c%d", 
+					 (minor >> 4) + 'a', (minor & 0xf)); 
+			else
+				sprintf (linkto, "/dev/sd%c", 
+					 (minor >> 4) + 'a'); 
+			
+			break;
+		 case SCSI_CDROM_MAJOR:
+			sprintf (linkto, "/dev/sr%d", minor);
+			break;
+		 case SCSI_TAPE_MAJOR:
+			if (minor & 0x80)
+				sprintf (linkto, "/dev/nst%d", minor & 0x7f);
+			else
+				sprintf (linkto, "/dev/st%d", minor);
+			break;
+		 case SCSI_GENERIC_MAJOR:
+			sprintf (linkto, "/dev/sg%d", minor);
+			break;
+		}
+		unlink (scsidev);
+		symlink (linkto, scsidev);
+		
+		/*
+		 * Now make sure that the device the symlink points to actually
+		 * exists.  If not, then create that device.
+		 */
+		status = stat (scsidev, &statbuf);
+		if (status == 0)
+		{
+			if( statbuf.st_rdev != makedev(major, minor) )
+			{
+				fprintf(stderr,
+					"Warning - device %s does not point to expected device\n", linkto);
+			}
+		}
+		else
+		{
+			/*
+			 * Unable to stat the file.  Assume this is because it did not
+			 * exist, so we create it.
+			 */
+			newmode = devtype | filemode;
+			status = mknod(linkto, newmode,
+				       makedev(major,minor));
+			fprintf(stderr,"Creating %s\n", linkto);
+		}
+	}
+	else
+	{
+		update_device (scsidev, devtype, major, minor);
+	}
+}
 
-  status = stat(DEVSCSI, &statbuf);
-  if(status == -1)
+void build_disk (struct regnames * rpnt, int no)
+{
+	char scsidev[128]; 
+	int minor; int fd;
+	struct regnames * rpnt1 = regnames_dup (rpnt);
+	rpnt1->major = (no < 16? SCSI_DISK0_MAJOR: SCSI_DISK1_MAJOR + no/16 - 1);
+	rpnt1->minor = (no << 4) % 256;
+	sprintf (scsidev, "sdh%d-%xc%di%dl%d",
+		 rpnt1->hostnum, rpnt1->hostid,
+		 rpnt1->chan, rpnt1->id, rpnt1->lun);
+	rpnt1->name = strdup (scsidev);
+	rpnt1->next = reglist; reglist = rpnt1;
+	create_dev (rpnt1);
+	/* Check if device is there (i.e. medium inside) */
+	strcpy (scsidev, DEVSCSI); strcat (scsidev, "/"); 
+	strcat (scsidev, rpnt1->name);
+	fd =  open (scsidev, O_RDONLY);
+	if (fd < 0) return; /* No access to medium / part. table */
+	close (fd);
+	/* Now do a partition scan ... */
+	for (minor = rpnt1->minor+1; minor % 16; minor++)
+	{
+		struct stat statbuf; int status;
+		status = stat (TESTDEV, &statbuf);
+	if (status == 0)
+			unlink (TESTDEV);
+		
+		status = mknod ( TESTDEV, 0600 | S_IFBLK,
+				makedev (rpnt1->major, minor) );
+		fd = open (TESTDEV, O_RDONLY);
+		unlink (TESTDEV);
+	if (fd < 0) continue;
+		// TO DO: Add sanity checks here
+		close (fd);
+		rpnt1 = regnames_dup (rpnt1);
+		rpnt1->partition = minor % 16;
+		rpnt1->minor = minor;
+		sprintf (scsidev, "sdh%d-%xc%di%dl%dp%d",
+			 rpnt1->hostnum, rpnt1->hostid,
+			 rpnt1->chan, rpnt1->id, rpnt1->lun, rpnt1->partition);
+		rpnt1->name = strdup (scsidev);
+		rpnt1->next = reglist; reglist = rpnt1;
+		create_dev (rpnt1);
+	}
+}
+
+void build_tape (struct regnames * rpnt, int no)
+{
+	char scsidev[128];
+	struct regnames * rpnt1 = regnames_dup (rpnt);
+	rpnt1->major = SCSI_TAPE_MAJOR;
+	rpnt1->minor = no;
+	sprintf (scsidev, "sth%d-%xc%di%dl%d",
+		 rpnt1->hostnum, rpnt1->hostid,
+		 rpnt1->chan, rpnt1->id, rpnt1->lun);
+	rpnt1->name = strdup (scsidev);
+	rpnt1->next = reglist; reglist = rpnt1;
+	create_dev (rpnt1);
+	/* And create the no-rewind alias */
+	rpnt1 = regnames_dup (rpnt1);
+	rpnt1->minor |= 0x80;
+	sprintf (scsidev, "nsth%d-%xc%di%dl%d",
+		 rpnt1->hostnum, rpnt1->hostid,
+		 rpnt1->chan, rpnt1->id, rpnt1->lun);
+	rpnt1->name = strdup (scsidev);
+	rpnt1->next = reglist; reglist = rpnt1;
+	create_dev (rpnt1);
+}
+
+void build_cdrom (struct regnames * rpnt, int no)
+{
+   char scsidev[128];
+	struct regnames * rpnt1 = regnames_dup (rpnt);
+	rpnt1->major = SCSI_CDROM_MAJOR;
+	rpnt1->minor = no;
+	sprintf (scsidev, "srh%d-%xc%di%dl%d",
+		 rpnt1->hostnum, rpnt1->hostid,
+		 rpnt1->chan, rpnt1->id, rpnt1->lun);
+	rpnt1->name = strdup (scsidev);
+	rpnt1->next = reglist; reglist = rpnt1;
+	create_dev (rpnt1);
+}
+
+void build_sgdevlist ()
+{
+  int fd; 
+  char hostname[64];
+  char scsidev[128];
+  int h_id, id[2];
+  int scsi_id, lun;
+  int chan;
+  struct stat statbuf;
+  int status;
+  struct regnames * rpnt;
+  int disks = 0, tapes = 0, cdroms = 0;
+  int miss = 0;	
+  int minor = 0;
+  int major = SCSI_GENERIC_MAJOR; 
+  char devchar = 'g'; int mode = O_RDWR;
+  int devtype = (SCSI_BLK_MAJOR(major)? S_IFBLK: S_IFCHR);
+
+  status = stat (DEVSCSI, &statbuf);
+  if (status == -1)
       return;
 
-  status = stat(TESTDEV, &statbuf);
-  if(status == 0)
-    {
-      unlink(TESTDEV);
-    }
+  status = stat (TESTDEV, &statbuf);
+  if (status == 0)
+      unlink (TESTDEV);
 
-  if (verbose >=1)
+  if (verbose >= 1)
   	fprintf (stderr, "Building list for s%c (%s dev major %i)\n",
-		devchar, (devtype==S_IFCHR? "char": "block"), major);
-  while(1==1)
+		 devchar, (SCSI_BLK_MAJOR(major)? "block": "char"), major);
+  while (minor <= 255)
     {
-      status = mknod(TESTDEV,0600 | devtype ,
-		     makedev(major,minor));
-      fd = open(TESTDEV, mode);
-      if(fd == -1) {
-	if( verbose == 2 )
+      errno = 0;
+      status = mknod ( TESTDEV, 0600 | devtype ,
+		       makedev (major,minor) );
+      if (status) { perror ("scsidev: mknod"); exit (3); }
+      fd = open (TESTDEV, mode);
+      unlink (TESTDEV);
+      if (fd == -1) {
+	if (verbose == 2)
 	  {
-	    fprintf(stderr,"open(%x/%x) returned %d (%d)\n",
-		    major, minor, fd, errno);
+	    fprintf (stderr, "open(%x/%x) returned %d (%d)\n",
+		     major, minor, fd, errno);
 	  }
-	/*
-	 * For disks, only stop searching when we cannot find a whole-disk
-	 * somewhere (there could be sparse partition tables).  For anything
-	 * else, stop as soon as we find something non-existant.
-	 */
-	if(   (major == SCSI_DISK_MAJOR)
-	      && minor < 255)
-	  {
-	    minor++;
-	    unlink (TESTDEV);
-	    continue;
-	  } else if ( major != SCSI_DISK_MAJOR )
-	    break;
+	miss++;
+	if (miss > maxmiss) break;
+	  else { minor++; continue; }
       }
-      status = ioctl(fd, SCSI_IOCTL_GET_IDLUN, &id);
+      status = ioctl (fd, SCSI_IOCTL_GET_IDLUN, &id);
 
-      if(status == -1)	{
-	if( verbose == 2 )
+      if (status == -1)	{
+	if (verbose == 2)
 	  {
-	    fprintf(stderr,"idlun(%x/%x) returned %d (%d)\n",
-		    major, minor, status, errno);
+	    fprintf (stderr, "idlun(%x/%x) returned %d (%d)\n",
+		     major, minor, status, errno);
 	  }
-	close(fd);
+	close (fd);
 	break;
       }
 
-      status = ioctl(fd, SCSI_IOCTL_PROBE_HOST, hostname);
-      close(fd);
+      *(int*)hostname = 63; // bufflen, used by ioctl_probe (), KG, 2000-01-27
+      status = ioctl (fd, SCSI_IOCTL_PROBE_HOST, hostname);
+      close (fd);
 
-      if(status == -1)	{
-	if( verbose == 2 )
+      if (status == -1)	{
+	if (verbose == 2)
 	  {
-	    fprintf(stderr,"probe(%x/%x) returned %d (%d)\n",
-		    major, minor, status, errno);
+	    fprintf (stderr, "probe(%x/%x) returned %d (%d)\n",
+		     major, minor, status, errno);
 	  }
 	break;
       };
 
       if (verbose == 2)
-	 fprintf(stderr, "Found scsi device %x with idlun %x\n", makedev(major, minor), id[0]);
+	 fprintf (stderr, "Found %x with idlun %x\n", makedev(major, minor), id[0]);
       h_id = id[0] >> 24;
       chan = (id[0] >> 16) & 0xff;
       lun = (id[0] >> 8 ) & 0xff;
       scsi_id = id[0] & 0xff;
-      if(major == SCSI_DISK_MAJOR && (minor & 0xf))
-	{
-	  sprintf(scsidev,DEVSCSI "/s%ch%d-%xc%di%dl%dp%d",
-		  devchar,
-		  h_id,
-		  id[1],
-		  chan,
-		  scsi_id,
-		  lun, (minor & 0xf));
-	  register_dev(scsidev, major, minor, 
-		       h_id, id[1], chan, scsi_id, 
-		       lun, (minor & 0xf), NULL);
-      } else {
-	sprintf(scsidev,DEVSCSI "/s%ch%d-%xc%di%dl%d",
-		devchar,
-		h_id,
-		id[1],
-		chan,
-		scsi_id,
-		lun);
-	register_dev(scsidev, major, minor, 
-		     h_id, id[1], chan, scsi_id, lun, -1, NULL);
+      sprintf (scsidev, DEVSCSI "/s%ch%d-%xc%di%dl%d",
+		devchar, h_id, id[1],
+		chan, scsi_id, lun);
+      rpnt = register_dev (scsidev, major, minor, 
+			   h_id, id[1], chan, scsi_id, lun, -1, hostname, NULL);
+      create_dev (rpnt);
+      inquiry (rpnt);
+      if (!quiet) printf ("Found %s (Type %02x) %c on %s \n", scsidev,
+			  rpnt->inq_devtp, (rpnt->rmvbl? 'R' : ' '),
+			  hostname);
+
+      /* Now register cdroms, tapes, and disks as well */
+      switch (rpnt->inq_devtp) {
+       case TYPE_DISK:
+       case TYPE_MOD:
+	 	build_disk (rpnt, disks++);
+		break;
+       case TYPE_TAPE: 
+	 	build_tape (rpnt, tapes++);
+		break;
+       case TYPE_ROM:
+       case TYPE_WORM:
+	 	build_cdrom (rpnt, cdroms++);
+	 	break;
+       default:
+	 /* nothing to be done */
       }
-      printf("Found scsi device %s %s\n", scsidev, hostname);
-
-      if( use_symlink )
-	{
-	  /*
-	   * Figure out what we should be symlinked to here.
-	   */
-	  switch(major)
-	    {
-	    case SCSI_DISK_MAJOR:
-	      if( minor & 0xf)
-		sprintf(linkto, "/dev/sd%c%d", 
-			(minor >> 4) + 'a', (minor & 0xf)); 
-	      else
-		sprintf(linkto, "/dev/sd%c", 
-			(minor >> 4) + 'a'); 
-
-	      break;
-	    case SCSI_CDROM_MAJOR:
-	      sprintf(linkto, "/dev/sr%d", minor);
-	      break;
-	    case SCSI_TAPE_MAJOR:
-	      sprintf(linkto, "/dev/st%d", minor);
-	      break;
-	    case SCSI_GENERIC_MAJOR:
-	      sprintf(linkto, "/dev/sg%d", minor);
-	      break;
-	    }
-
-	  symlink(linkto, scsidev);
-
-	  /*
-	   * Now make sure that the device the symlink points to actually
-	   * exists.  If not, then create that device.
-	   */
-	  status = stat(scsidev, &statbuf);
-	  if( status == 0 )
-	    {
-	      if( statbuf.st_rdev != makedev(major, minor) )
-		{
-		  fprintf(stderr,
-			  "Warning - device %s does not point to expected device\n", linkto);
-		}
-	    }
-	  else
-	    {
-	      /*
-	       * Unable to stat the file.  Assume this is because it did not
-	       * exist, so we create it.
-	       */
-	      newmode = devtype | 
-		(major != SCSI_CDROM_MAJOR ? filemode : (filemode & ~0222));
-	      status = mknod(linkto, newmode,
-			     makedev(major,minor));
-	      fprintf(stderr,"Creating %s\n", linkto);
-	    }
-	  
-	  if( major == SCSI_TAPE_MAJOR )
-	    {
-	      sprintf(scsidev,DEVSCSI "/ns%ch%d-%xc%di%dl%d",
-		      devchar,
-		      h_id,
-		      id[1],
-		      chan,
-		      scsi_id,
-		      lun);
-	      sprintf(linkto, "/dev/nst%d", minor);
-	      symlink(linkto, scsidev);
-	      /*
-	       * Now make sure that the device the symlink points to actually
-	       * exists.  If not, then create that device.
-	       */
-	      status = stat(scsidev, &statbuf);
-	      if( status == 0 )
-		{
-		  if( statbuf.st_rdev != makedev(major, minor) )
-		    {
-		      fprintf(stderr,
-			      "Warning - device %s does not point to expected device\n", linkto);
-		    }
-		}
-	      else
-		{
-		  /*
-		   * Unable to stat the file.  Assume this is because it did not
-		   * exist, so we create it.
-		   */
-		  newmode = devtype | 
-		    (major != SCSI_CDROM_MAJOR ? filemode : (filemode & ~0222));
-		  status = mknod(linkto, newmode,
-				 makedev(major,minor));
-		  fprintf(stderr,"Creating %s\n", linkto);
-		}
-	    }
-	}
-      else
-	{
-	  update_device(scsidev, devtype, major, minor);
-
-	  if( major == SCSI_TAPE_MAJOR )
-	    {
-	      sprintf(scsidev,DEVSCSI "/ns%ch%d-%xc%di%dl%d",
-		      devchar,
-		      h_id,
-		      id[1],
-		      chan,
-		      scsi_id,
-		      lun);
-	      register_dev(scsidev, major, minor | 0x80, 
-			   h_id, id[1], chan, scsi_id, lun, -1, NULL);
-
-	      update_device(scsidev, devtype, major, minor | 0x80);
-	    }
-	}
-      
-      unlink(TESTDEV);
       minor += 1;
     }
-  unlink(TESTDEV);
+  unlink (TESTDEV);
 }
 
 
-int get_serial_number(struct regnames *);
-void build_special();
-int inquire(struct regnames *);
-
-/*
- * Build a /dev/scsi tree that contains the correct device entries.
- */
-
 void usage()
 {
-  fprintf(stderr, "%s\n", versinfo);
-  fprintf(stderr,"Usage: scsidev [-f] [-l] [-v] [-m mode]\n");
+	fprintf (stderr, "%s\n", versid);
+	fprintf (stderr, "Usage: scsidev [options]\n");
+	fprintf (stderr, " -f     : force cleaning of " DEVSCSI" dir\n");
+	fprintf (stderr, " -l/-L  : create symlinks for device names / alias names\n");
+	fprintf (stderr, " -v/-q  : verbose/quiet operation\n");
+	fprintf (stderr, " -m mode: permission to create dev nodes with\n");
+	fprintf (stderr, " -s     : list serial numbers of devices (if available)\n");
+	fprintf (stderr, " -c mxms: continue scanning until mxms missing devs found\n");
 }
 
-int main(int argc, char * argv[])
+int main (int argc, char * argv[])
 {
   char c;
   int force = 0;
@@ -558,51 +654,49 @@ int main(int argc, char * argv[])
       fprintf(stderr, DEVSCSI " either does not exist, or is not a directory\n");
       exit(0);
     }
-  while ((c = getopt(argc, argv, "flvsm:")) != EOF)
+  while ((c = getopt(argc, argv, "flLvqshm:c:")) != EOF)
     {
       switch (c)
 	{
 	case 'f':
-	  force = 1;
-	  break;
+	  force = 1; break;
 	case 'm':
-	  filemode = strtoul(optarg, 0, 0);
-	  break;
+	  filemode = strtoul (optarg, 0, 0); break;
+	case 'c':
+	  maxmiss = strtoul (optarg, 0, 0); break;
 	case 'l':
-	  use_symlink = 1;
-	  break;
+	  use_symlink = 1; break;
+	case 'L':
+	  symlink_alias = 1; break;
 	case 's':
-	  show_serial = 1;
-	  break;
+	  show_serial = 1; break;
 	case 'v':
-	  verbose++;
-	  break;
+	  verbose++; break;
+	case 'h':
+	  usage (); exit (0); break;
+	case 'q':
+	  quiet = 1; break;
 	default:
-	  usage();
-	  exit(1);
-	  break;
+	  usage(); exit(1); break;
 	}
     }
 
-  if( verbose >= 1 ) fprintf( stderr, "%s\n", versinfo );
+  if( verbose >= 1 ) fprintf( stderr, "%s\n", versid );
   if( force ) flush_sdev();
 
 #ifdef DEBUG
-  register_dev("/dev/scsi/sdh4-334c0i0l0",   8,  0, 6, 0x334, 0, 0, 0, -1, NULL);
-  register_dev("/dev/scsi/sdh4-334c0i0l0p1", 8,  1, 6, 0x334, 0, 0, 0,  1, NULL);
-  register_dev("/dev/scsi/sdh4-334c0i0l0p2",8,  2,6, 0x334, 0, 0, 0,  2, NULL);
-  register_dev("/dev/scsi/sdh4-334c0i0l0p3",8,  3,6, 0x334, 0, 0, 0,  3, NULL);
-  register_dev("/dev/scsi/sgh4-334c0i0l0", 21,  0,6, 0x334, 0, 0, 0, -1, NULL);
-  register_dev("/dev/scsi/sgh4-334c0i2l0", 21,  1,6, 0x334, 0, 2, 0, -1, NULL);
-  register_dev("/dev/scsi/sgh4-334c0i5l0", 21,  2,6, 0x334, 0, 5, 0, -1, NULL);
-  register_dev("/dev/scsi/srh4-334c0i2l0", 11,  0,6, 0x334, 0, 2, 0, -1, NULL);
-  register_dev("/dev/scsi/sth4-334c0i5l0",  9,  0,6, 0x334, 0, 5, 0, -1, NULL);
-  register_dev("/dev/scsi/rsth4-334c0i5l0", 9,128,6, 0x334, 0, 5, 0, -1, NULL);
+  register_dev("/dev/scsi/sdh4-334c0i0l0",   8,  0, 6, 0x334, 0, 0, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/sdh4-334c0i0l0p1", 8,  1, 6, 0x334, 0, 0, 0,  1, "debug", NULL);
+  register_dev("/dev/scsi/sdh4-334c0i0l0p2",8,  2,6, 0x334, 0, 0, 0,  2, "debug", NULL);
+  register_dev("/dev/scsi/sdh4-334c0i0l0p3",8,  3,6, 0x334, 0, 0, 0,  3, "debug", NULL);
+  register_dev("/dev/scsi/sgh4-334c0i0l0", 21,  0,6, 0x334, 0, 0, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/sgh4-334c0i2l0", 21,  1,6, 0x334, 0, 2, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/sgh4-334c0i5l0", 21,  2,6, 0x334, 0, 5, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/srh4-334c0i2l0", 11,  0,6, 0x334, 0, 2, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/sth4-334c0i5l0",  9,  0,6, 0x334, 0, 5, 0, -1, "debug", NULL);
+  register_dev("/dev/scsi/rsth4-334c0i5l0", 9,128,6, 0x334, 0, 5, 0, -1, "debug", NULL);
 #else
-  build_sdev(SCSI_DISK_MAJOR,    O_RDONLY, S_IFBLK, 'd');
-  build_sdev(SCSI_CDROM_MAJOR,   O_RDONLY, S_IFBLK, 'r');
-  build_sdev(SCSI_TAPE_MAJOR,    O_RDONLY, S_IFCHR, 't');
-  build_sdev(SCSI_GENERIC_MAJOR, O_RDWR,   S_IFCHR, 'g');
+  build_sgdevlist ();
 #endif
 
   if( show_serial )
@@ -613,18 +707,20 @@ int main(int argc, char * argv[])
 	  if( rpnt->major == SCSI_TAPE_MAJOR
 	      && (rpnt->minor & 0x80) != 0 ) continue;
 
-	  if( rpnt->serial == NULL )  get_serial_number(rpnt);
+	  //if( rpnt->serial == NULL ) inquiry (rpnt);
 	  if( rpnt->serial == no_serial )
 	    printf("Device  %s has no serial number\n", rpnt->name);
 	  else
 	    printf("Serial number of %s: \"%s\"\n", rpnt->name, rpnt->serial);
 	}
     }
+
+  //dumplist ();
   /*
    * Now, read the configuration file and see whether there
    * are any special device names we want to try and match.
    */
-  build_special();
+  build_special ();
 
   if( !force )
     {
@@ -732,11 +828,12 @@ char * get_number(char * pnt, int * result)
  * LUN=number
  * CHANNEL=number
  * HOSTID=number
+ * HOST="string" (hostname)
  * MANUFACTURER="string"
  * MODEL="string"
- * REV="string"
  * SERIAL_NUMBER="string"	(for those devices that support this).
- * NAME="string"
+ * REV="string"
+ * NAME="string" (alias)
  * DEVTYPE="disk", "tape", "generic", or "cdrom".
  */
 void build_special()
@@ -752,7 +849,7 @@ void build_special()
   int lun, chan, id, part, hostid, hostnum;
   int line;
   int devtype_i;
-  char *manufacturer, *model, *serial_number, *name, *devtype;
+  char *manufacturer, *model, *serial_number, *name, *devtype, *rev, *host;
   char *scsialias;
 
 #ifdef DEBUG
@@ -766,13 +863,12 @@ void build_special()
 	  if (verbose) perror (scsialias);
 	  return;
   }
-
   line = 0;
-  while(1==1)
+  while (1==1)
     {
       fgets (buffer, sizeof(buffer), configfile);
       line++;
-      if( feof(configfile)) break;
+      if (feof (configfile)) break;
 
       /*
        * Remove trailing \n, if present.
@@ -783,12 +879,13 @@ void build_special()
       /*
        * First, tokenize the input line, and pick out the parameters.
        */
+      lun = -1; id = -1;
       chan = -1;
-      id = -1; lun = -1;
-      part = -1;
       hostid = -1; hostnum = -1;
+      part = -1;
+      host = NULL;
       manufacturer = NULL; model = NULL;
-      serial_number = NULL;
+      serial_number = NULL; rev = NULL;
       name = NULL;
       devtype = NULL;
       pnt = buffer;
@@ -817,6 +914,14 @@ void build_special()
 	    {
 	      pnt = get_string(pnt1 + 1, &serial_number);
 	    }
+	  else if ( strncmp(pnt, "rev", 3) == 0 )
+	    {
+	      pnt = get_string(pnt1 + 1, &rev);
+	    }
+	  else if ( strncmp(pnt, "hostname", 6) == 0 )
+	    {
+	      pnt = get_string(pnt1 + 1, &host);
+	    }
 	  else if ( strcmp(pnt, "id") == 0 )
 	    {
 	      pnt = get_number(pnt1 + 1, &id);
@@ -893,37 +998,28 @@ void build_special()
       match = NULL;
       for (rpnt = reglist; rpnt; rpnt = rpnt->next)
 	{
+	  if( rpnt->alias != NULL ) continue;
 	  /*
 	   * Check the integers first.  Some of the strings we have to
 	   * request, and we want to avoid this if possible.
 	   */
-	  if( rpnt->alias != NULL ) continue;
 	  if( id != -1 && id != rpnt->id ) continue;
 	  if( chan != -1 && id != rpnt->chan ) continue;
 	  if( lun != -1 && lun != rpnt->lun ) continue;
 	  if( hostid != -1 && hostid != rpnt->hostid ) continue;
 	  if( hostnum != -1 && hostnum != rpnt->hostnum ) continue;
-	  if( devtype_i != rpnt->major ) continue;
+	  if( devtype_i != rpnt->major && 
+	      (devtype_i != SCSI_DISK0_MAJOR || ! SCSI_DISK_MAJOR(rpnt->major)) )
+		continue;
 	  if( part != rpnt->partition ) continue;
-	  if(    rpnt->major == SCSI_TAPE_MAJOR 
+	  if( rpnt->major == SCSI_TAPE_MAJOR 
 	      && (rpnt->minor & 0x80) != 0) continue;
 
 	  /*
 	   * OK, that matches, now obtain some of the strings
 	   * that might be needed.
 	   */
-	  if(    (manufacturer != NULL || model != NULL)
-	      && rpnt->manufacturer == NULL)
-	    {
-	      inquire(rpnt);
-	    }
-	  
-	  if( serial_number && rpnt->serial == NULL )
-	    {
-	      get_serial_number(rpnt);
-	    }
-
-	  if(    manufacturer != NULL && (rpnt->manufacturer == NULL ||
+	  if( manufacturer != NULL && (rpnt->manufacturer == NULL ||
 	      	strcmp(rpnt->manufacturer, manufacturer) != 0 ))
 	    continue;
 
@@ -935,31 +1031,37 @@ void build_special()
 	      	strcmp(rpnt->serial, serial_number) != 0 ))
 	    continue;
 
+  	  if( rev != NULL && (rpnt->rev == NULL ||
+	      	strcmp(rpnt->rev, rev) != 0 ))
+	    continue;
+
+     	  if( host != NULL && (rpnt->hostname == NULL ||
+	      	strncmp(rpnt->hostname, host, strlen(host)) != 0 ))
+	    continue;
+
 	  /*
 	   * We have a match.  Record it and keep looking just in
 	   * case we find a duplicate.
 	   */
 	  if( match != NULL )
 	    {
-	      fprintf(stderr,"Line %d not matched uniquely\n", line);
-	      fprintf(stderr," Prev. match: %s\n", match->name);
-	      fprintf(stderr," Curr. match: %s\n", rpnt->name);
+	      fprintf (stderr, "Line %d not matched uniquely\n", line);
+	      fprintf (stderr, " Prev. match: %s\n", match->name);
+	      fprintf (stderr, " Curr. match: %s\n", rpnt->name);
 	      break;
 	    }
 	  else
-	    {
 	      match = rpnt;
-	    }
 	}
 
       /*
        * See if there was a non-unique mapping.  If so, then
        * don't do anything for this one.
        */
+
+      // detect break
       if( rpnt != NULL )
-	{
 	  continue;
-	}
 
       if( match != NULL )
 	{
@@ -967,9 +1069,9 @@ void build_special()
 	   * OK, we have a unique match.  Create the device entries,
 	   * as requested.
 	   */
-	  fprintf(stderr,"Building alias device entries for %s...\n", name);
+	  if (!quiet) fprintf (stderr, "Building alias device entries for %s...\n", name);
 
-	  if( devtype_i == SCSI_DISK_MAJOR
+	  if( devtype_i == SCSI_DISK0_MAJOR
 	      || devtype_i == SCSI_CDROM_MAJOR )
 	    type = S_IFBLK;
 	  else
@@ -979,22 +1081,32 @@ void build_special()
 	   * If this is just an ordinary single device type,
 	   * Just create it.
 	   */
-	  sprintf(scsidev, DEVSCSI "/%s", name);
-	  update_device(scsidev, type, 
-			match->major, match->minor);
+	  sprintf (scsidev, DEVSCSI "/%s", name);
+	  unlink (scsidev);
+	  if (symlink_alias)
+		symlink (match->name, scsidev);
+	  else
+		update_device(scsidev, type, 
+			      match->major, match->minor);
 	  register_dev(scsidev, match->major, match->minor | 0x80, 
-		       0, 0, 0, 0, 0, 0, match);
+		       0, 0, 0, 0, 0, 0, 0, match);
 
 	  if( devtype_i == SCSI_TAPE_MAJOR )
 	    {
-	      sprintf(scsidev, DEVSCSI "/n%s", name);
-	      update_device(scsidev, S_IFCHR, 
-			    match->major, (match->minor | 0x80) );
+	      sprintf (scsidev, DEVSCSI "/n%s", name);
+	      unlink (scsidev);
+	      if (symlink_alias) {
+		char nm2[128]; strcpy (nm2, "n"); strcat (nm2, match->name);
+		symlink (nm2, scsidev);
+	      }
+	      else
+		update_device(scsidev, S_IFCHR,
+			      match->major, (match->minor | 0x80) );
 	      register_dev(scsidev, match->major, match->minor | 0x80, 
-			   0, 0, 0, 0, 0, 0, match);
+			   0, 0, 0, 0, 0, 0, 0, match);
 	    }
 
-	  if(    devtype_i == SCSI_DISK_MAJOR
+	  if ( devtype_i == SCSI_DISK0_MAJOR
 		 && match->partition == -1 )
 	    {
 	      /*
@@ -1017,23 +1129,40 @@ void build_special()
 
 		  sprintf(scsidev, DEVSCSI "/%s-p%d", name, 
 			  rpnt->partition);
-		  update_device(scsidev, S_IFBLK, 
-				match->major, rpnt->minor);
+		  unlink (scsidev);
+		  if (symlink_alias)
+		     symlink (match->name, scsidev);
+		  else
+		     update_device(scsidev, S_IFBLK, 
+				   match->major, rpnt->minor);
 		  register_dev(scsidev, match->major, match->minor | 0x80, 
-			       0, 0, 0, 0, 0, 0, rpnt);
+			       0, 0, 0, 0, 0, 0, 0, rpnt);
 		}
 	    }
 	}
       else
 	{
-	  fprintf(stderr,"Unable to match device for line %d (alias %s)\n", 
-		  line, name);
+	  if (!quiet) fprintf (stderr, "Unable to match device for line %d (alias %s)\n", 
+		   line, name);
 	}
     }
 
-  fclose(configfile);
+  fclose (configfile);
 }
 
+void dumppage (unsigned char* page)
+{
+   int ln = 4 + page[4];
+   int i;
+   for (i = 0; i <= ln; i++)
+     {
+	printf (" %02x", page[i]);
+	if (!((i+1)%16)) printf ("\n");
+     }
+   if (i%16) printf ("\n");
+}
+
+
 char* getstr (char* page, int start, int stop)
 {
    int ln = stop-start+1; int i;
@@ -1044,72 +1173,7 @@ char* getstr (char* page, int start, int stop)
    return str;
 }
 
-/*
- * Note that this does the same thing as an inquiry, but
- * it also grabs the serial number too.  Some devices complain
- * if you ask for the serial number, so we have a version with and
- * without.
- */
-int get_serial_number(struct regnames * rpnt)
-{
-#ifdef DEBUG
-  /*
-   * Fill in some entries, just so I can test this on a non-scsi system.
-   */
-  if(rpnt->id == 0 )  rpnt->serial = strdup("DX908FK");
-  else  rpnt->serial = no_serial;
-#else
-  int status;
-  unsigned char *cmd;
-  unsigned char * pagestart;
-  unsigned char buffer[1024];
-  char path[64];
-  int infile;
-  struct regnames * rpnt1;
-
-  memset(buffer, 0, sizeof(buffer));
-  
-  *( (int *)  buffer )		= 0;	/* length of input data */
-  *( ((int *) buffer) + 1 )	= 512;	/* length of output buffer */
-
-  cmd = (char *) ( ((int *) buffer) + 2 );
-  
-  cmd[0] = 0x12;			/* INQUIRY */
-  cmd[1] = 0x01;			/* lun=0, evpd=1 */
-  cmd[2] = 0x80;			/* page code = 0x80, serial number */
-  cmd[3] = 0x00;			/* (reserved) */
-  cmd[4] = 0xff;			/* allocation length */
-  cmd[5] = 0x00;			/* control */
-
-  strcpy(path, DEVSCSI); strcat(path, "/");
-  strcat(path, rpnt->name);
-  infile = open(path, O_RDWR);
-  if( infile == -1 ) return -1;
-  status = ioctl( infile, 1 /* SCSI_IOCTL_SEND_COMMAND */, buffer );
-  close(infile);
-
-  pagestart = buffer + 8;
-  if(status) rpnt->serial = no_serial;
-  else rpnt->serial = getstr (pagestart, 4, 3+pagestart[3]);
-
-  /*
-   * Copy information to identical nodes to avoid excessive probing.
-   */
-  for( rpnt1 = reglist; rpnt1; rpnt1 = rpnt1->next )
-    {
-      if( rpnt1 == rpnt ) continue;
-      if( rpnt1->alias != NULL ) continue;
-      if( rpnt1->id != rpnt->id ) continue;
-      if( rpnt1->chan != rpnt->chan ) continue;
-      if( rpnt1->hostnum != rpnt->hostnum ) continue;
-      if( rpnt1->hostid != rpnt->hostid ) continue;
-      if( rpnt1->serial == NULL ) rpnt1->serial = rpnt->serial;
-    }
-   return 0;
-#endif
-}
-
-int inquire(struct regnames * rpnt)
+int inquiry (struct regnames * rpnt)
 {
 #ifdef DEBUG
   /*
@@ -1131,10 +1195,9 @@ int inquire(struct regnames * rpnt)
   unsigned char * pagestart;
   unsigned char buffer[1024];
   char path[64];
-  struct regnames * rpnt1;
   int infile;
 
-  memset(buffer, 0, sizeof(buffer));
+  memset (buffer, 0, sizeof(buffer));
   
   *( (int *)  buffer )		= 0;	/* length of input data */
   *( ((int *) buffer) + 1 )	= sizeof(buffer)-16;	/* length of output buffer */
@@ -1142,8 +1205,8 @@ int inquire(struct regnames * rpnt)
   cmd = (char *) ( ((int *) buffer) + 2 );
   
   cmd[0] = 0x12;			/* INQUIRY */
-  cmd[1] = 0x00;			/* lun=0, evpd=0 */
-  cmd[2] = 0x00;			/* page code = 0 */
+  cmd[1] = (rpnt->lun << 5) | 0x01;	/* lun, evpd=1 */
+  cmd[2] = 0x80;			/* page code = 0x80, serial number */
   cmd[3] = 0x00;			/* (reserved) */
   cmd[4] = 0xff;			/* allocation length */
   cmd[5] = 0x00;			/* control */
@@ -1151,38 +1214,53 @@ int inquire(struct regnames * rpnt)
   strcpy(path, DEVSCSI); strcat(path, "/");
   strcat(path, rpnt->name);
   infile = open(path, O_RDWR);
-  if( infile == -1 ) return 1;
+  if( infile == -1 ) return -1;
 
   status = ioctl( infile, 1, buffer );
-
   close(infile);
-
-  if(status) return 1;
-
   pagestart = buffer + 8;
 
-  rpnt->manufacturer = getstr(pagestart, 8, 15);
-  rpnt->model = getstr(pagestart, 16, 31);
-  rpnt->rev = getstr(pagestart, 32, 35);
+  if (status) 
+	rpnt->serial = no_serial;
+  else
+	rpnt->serial = getstr (pagestart, 4, 3+pagestart[3]);
+
+  if (verbose == 2) printf ("Serial for %s: %s\n", path, rpnt->serial);
+
+  memset (buffer, 0, sizeof(buffer));
+  
+  *( (int *)  buffer )		= 0;	/* length of input data */
+  *( ((int *) buffer) + 1 )	= sizeof(buffer)-16;	/* length of output buffer */
+
+  cmd = (char *) ( ((int *) buffer) + 2 );
+  
+  cmd[0] = 0x12;			/* INQUIRY */
+  cmd[1] = (rpnt->lun << 5);		/* lun, evpd=0 */
+  cmd[2] = 0x00;			/* page code 0: std inquiry */
+  cmd[3] = 0x00;			/* (reserved) */
+  cmd[4] = 0xff;			/* allocation length */
+  cmd[5] = 0x00;			/* control */
+
+  infile = open(path, O_RDWR);
+  if( infile == -1 ) return -1;
+
+  status = ioctl( infile, 1, buffer );
+  close(infile);
+  pagestart = buffer + 8;
+
+  if (status) 
+     { 
+	fprintf (stderr, "INQUIRY failed for %s (%i-%i)!\n", 
+		 path, rpnt->id, rpnt->lun);
+	return -1; 
+     }
+  if (verbose == 2) dumppage (pagestart);
+  rpnt->manufacturer = getstr (pagestart, 8, 15); 
+  rpnt->model = getstr (pagestart, 16, 31);
+  rpnt->rev = getstr (pagestart, 32, 35);
+  rpnt->inq_devtp = pagestart[0] & 0x1f;
+  rpnt->rmvbl = (pagestart[1] & 0x80) >> 7;
 
-  /*
-   * Copy information to identical nodes to avoid excessive probing.
-   */
-  for( rpnt1 = reglist; rpnt1; rpnt1 = rpnt1->next )
-    {
-      if( rpnt1 == rpnt ) continue;
-      if( rpnt1->alias != NULL ) continue;
-      if( rpnt1->id != rpnt->id ) continue;
-      if( rpnt1->chan != rpnt->chan ) continue;
-      if( rpnt1->hostnum != rpnt->hostnum ) continue;
-      if( rpnt1->hostid != rpnt->hostid ) continue;
-      if( rpnt1->manufacturer == NULL ) 
-		rpnt1->manufacturer = rpnt->manufacturer;
-      if( rpnt1->model == NULL ) 
-	 	rpnt1->model = rpnt->model;
-      if( rpnt1->rev == NULL ) 
-	 	rpnt1->rev = rpnt->rev;
-    }
   return 0;
 #endif
 }
